@@ -3,14 +3,19 @@ use libc::{c_char, c_double};
 use num_traits::Float;
 use proj_sys::{
     proj_area_create, proj_area_destroy, proj_area_set_bbox, proj_cleanup, proj_context_create,
-    proj_context_destroy, proj_context_get_url_endpoint, proj_context_is_network_enabled,
-    proj_context_set_search_paths, proj_context_set_url_endpoint, proj_create,
-    proj_create_crs_to_crs, proj_destroy, proj_errno_string, proj_get_area_of_use,
+    proj_context_destroy, proj_context_errno, proj_context_get_url_endpoint,
+    proj_context_is_network_enabled, proj_context_set_search_paths, proj_context_set_url_endpoint,
+    proj_create, proj_create_crs_to_crs, proj_destroy, proj_errno_string, proj_get_area_of_use,
     proj_grid_cache_set_enable, proj_info, proj_normalize_for_visualization, proj_pj_info,
     proj_trans, proj_trans_array, PJconsts, PJ_AREA, PJ_CONTEXT, PJ_COORD, PJ_DIRECTION_PJ_FWD,
     PJ_DIRECTION_PJ_INV, PJ_INFO, PJ_LP, PJ_XY,
 };
-use std::fmt::{self, Debug};
+use std::{
+    convert,
+    ffi,
+    fmt::{self, Debug},
+    str,
+};
 
 #[cfg(feature = "network")]
 use proj_sys::proj_context_set_enable_network;
@@ -21,12 +26,35 @@ use std::ffi::CStr;
 use std::ffi::CString;
 use std::mem::MaybeUninit;
 use std::path::Path;
-use std::str;
 use thiserror::Error;
 
 pub trait CoordinateType: Float + Copy + PartialOrd + Debug {}
 impl<T: Float + Copy + PartialOrd + Debug> CoordinateType for T {}
 
+/// An error number returned from a PROJ call.
+pub(crate) struct Errno(pub libc::c_int);
+
+impl Errno {
+    /// Return the error message associated with the error number.
+    pub fn message(&self, context: *mut PJ_CONTEXT) -> String {
+        let ptr = unsafe { proj_sys::proj_context_errno_string(context, self.0) };
+        if ptr.is_null() {
+            panic!("PROJ did not supply an error")
+        } else {
+            unsafe { _string(ptr).expect("PROJ provided an invalid error string") }
+        }
+    }
+}
+
+/// Construct a `Result` from the result of a `proj_create*` call.
+fn result_from_create<T>(context: *mut PJ_CONTEXT, ptr: *mut T) -> Result<*mut T, Errno> {
+    if ptr.is_null() {
+        Err(Errno(unsafe { proj_context_errno(context) }))
+    } else {
+        Ok(ptr)
+    }
+}
+
 /// A point in two dimensional space. The primary unit of input/output for proj.
 ///
 /// By default, any numeric `(x, y)` tuple implements `Coord`, but you can conform your type to
@@ -96,6 +124,14 @@ pub enum ProjError {
     DownloadError(String, String, u8),
 }
 
+#[derive(Error, Debug)]
+pub enum ProjCreateError {
+    #[error("A nul byte was found in the PROJ string definition or CRS argument: {0}")]
+    ArgumentNulError(ffi::NulError),
+    #[error("The underlying PROJ call failed: {0}")]
+    ProjError(String),
+}
+
 /// The bounding box of an area of use
 ///
 /// In the case of an area of use crossing the antimeridian (longitude +/- 180 degrees),
@@ -124,14 +160,14 @@ impl Area {
 }
 
 /// Easily get a String from the external library
-pub(crate) unsafe fn _string(raw_ptr: *const c_char) -> Result<String, ProjError> {
+pub(crate) unsafe fn _string(raw_ptr: *const c_char) -> Result<String, str::Utf8Error> {
     assert!(!raw_ptr.is_null());
     let c_str = CStr::from_ptr(raw_ptr);
     Ok(str::from_utf8(c_str.to_bytes())?.to_string())
 }
 
 /// Look up an error message using the error code
-fn error_message(code: c_int) -> Result<String, ProjError> {
+fn error_message(code: c_int) -> Result<String, str::Utf8Error> {
     unsafe {
         let rv = proj_errno_string(code);
         _string(rv)
@@ -149,45 +185,46 @@ fn area_set_bbox(parea: *mut proj_sys::PJ_AREA, new_area: Option<Area>) {
 }
 
 /// called by Proj::new and ProjBuilder::transform_new_crs
-fn transform_string(ctx: *mut PJ_CONTEXT, definition: &str) -> Option<Proj> {
-    let c_definition = CString::new(definition).ok()?;
-    let new_c_proj = unsafe { proj_create(ctx, c_definition.as_ptr()) };
-    if new_c_proj.is_null() {
-        None
-    } else {
-        Some(Proj {
-            c_proj: new_c_proj,
-            ctx,
-            area: None,
-        })
-    }
+fn transform_string(ctx: *mut PJ_CONTEXT, definition: &str) -> Result<Proj, ProjCreateError> {
+    let c_definition =
+        CString::new(definition).map_err(|e| ProjCreateError::ArgumentNulError(e))?;
+    let ptr = result_from_create(ctx, unsafe { proj_create(ctx, c_definition.as_ptr()) })
+        .map_err(|e| ProjCreateError::ProjError(e.message(ctx)))?;
+    Ok(Proj {
+        c_proj: ptr,
+        ctx,
+        area: None,
+    })
 }
 
 /// Called by new_known_crs and proj_known_crs
-fn transform_epsg(ctx: *mut PJ_CONTEXT, from: &str, to: &str, area: Option<Area>) -> Option<Proj> {
-    let from_c = CString::new(from).ok()?;
-    let to_c = CString::new(to).ok()?;
+fn transform_epsg(
+    ctx: *mut PJ_CONTEXT,
+    from: &str,
+    to: &str,
+    area: Option<Area>,
+) -> Result<Proj, ProjCreateError> {
+    let from_c = CString::new(from).map_err(|e| ProjCreateError::ArgumentNulError(e))?;
+    let to_c = CString::new(to).map_err(|e| ProjCreateError::ArgumentNulError(e))?;
     let proj_area = unsafe { proj_area_create() };
     area_set_bbox(proj_area, area);
-    let new_c_proj =
-        unsafe { proj_create_crs_to_crs(ctx, from_c.as_ptr(), to_c.as_ptr(), proj_area) };
-    if new_c_proj.is_null() {
-        None
-    } else {
-        // Normalise input and output order to Lon, Lat / Easting Northing by inserting
-        // An axis swap operation if necessary
-        let normalised = unsafe {
-            let normalised = proj_normalize_for_visualization(ctx, new_c_proj);
-            // deallocate stale PJ pointer
-            proj_destroy(new_c_proj);
-            normalised
-        };
-        Some(Proj {
-            c_proj: normalised,
-            ctx,
-            area: Some(proj_area),
-        })
-    }
+    let ptr = result_from_create(ctx, unsafe {
+        proj_create_crs_to_crs(ctx, from_c.as_ptr(), to_c.as_ptr(), proj_area)
+    })
+    .map_err(|e| ProjCreateError::ProjError(e.message(ctx)))?;
+    // Normalise input and output order to Lon, Lat / Easting Northing by inserting
+    // An axis swap operation if necessary
+    let normalised = unsafe {
+        let normalised = proj_normalize_for_visualization(ctx, ptr);
+        // deallocate stale PJ pointer
+        proj_destroy(ptr);
+        normalised
+    };
+    Ok(Proj {
+        c_proj: normalised,
+        ctx,
+        area: Some(proj_area),
+    })
 }
 
 /// Read-only utility methods for providing information about the current PROJ instance
@@ -226,7 +263,7 @@ pub trait Info {
     /// # Safety
     /// This method contains unsafe code.
     fn get_url_endpoint(&self) -> Result<String, ProjError> {
-        unsafe { _string(proj_context_get_url_endpoint(self.ctx())) }
+        Ok(unsafe { _string(proj_context_get_url_endpoint(self.ctx()))? })
     }
 }
 
@@ -244,7 +281,7 @@ impl ProjBuilder {
     /// This method contains unsafe code.
     #[cfg_attr(docsrs, doc(cfg(feature = "network")))]
     #[cfg(feature = "network")]
-    pub fn enable_network(&self, enable: bool) -> Result<u8, ProjError> {
+    pub fn enable_network(&mut self, enable: bool) -> Result<u8, ProjError> {
         if enable {
             let _ = match crate::network::set_network_callbacks(self.ctx()) {
                 1 => Ok(1),
@@ -272,7 +309,7 @@ impl ProjBuilder {
     ///
     /// # Safety
     /// This method contains unsafe code.
-    pub fn set_search_paths<P: AsRef<Path>>(&self, newpath: P) -> Result<(), ProjError> {
+    pub fn set_search_paths<P: AsRef<Path>>(&mut self, newpath: P) -> Result<(), ProjError> {
         let existing = self.info()?.searchpath;
         let pathsep = if cfg!(windows) { ";" } else { ":" };
         let mut individual: Vec<&str> = existing.split(pathsep).collect();
@@ -301,7 +338,7 @@ impl ProjBuilder {
     ///
     /// # Safety
     /// This method contains unsafe code.
-    pub fn grid_cache_enable(&self, enable: bool) {
+    pub fn grid_cache_enable(&mut self, enable: bool) {
         let enable = if enable { 1 } else { 0 };
         let _ = unsafe { proj_grid_cache_set_enable(self.ctx(), enable) };
     }
@@ -310,7 +347,7 @@ impl ProjBuilder {
     ///
     /// # Safety
     /// This method contains unsafe code.
-    pub fn set_url_endpoint(&self, endpoint: &str) -> Result<(), ProjError> {
+    pub fn set_url_endpoint(&mut self, endpoint: &str) -> Result<(), ProjError> {
         let s = CString::new(endpoint)?;
         unsafe { proj_context_set_url_endpoint(self.ctx(), s.as_ptr()) };
         Ok(())
@@ -365,7 +402,7 @@ impl ProjBuilder {
     ///
     /// # Safety
     /// This method contains unsafe code.
-    pub fn proj(mut self, definition: &str) -> Option<Proj> {
+    pub fn proj(mut self, definition: &str) -> Result<Proj, ProjCreateError> {
         let ctx = unsafe { std::mem::replace(&mut self.ctx, proj_context_create()) };
         transform_string(ctx, definition)
     }
@@ -408,7 +445,12 @@ impl ProjBuilder {
     ///
     /// # Safety
     /// This method contains unsafe code.
-    pub fn proj_known_crs(mut self, from: &str, to: &str, area: Option<Area>) -> Option<Proj> {
+    pub fn proj_known_crs(
+        mut self,
+        from: &str,
+        to: &str,
+        area: Option<Area>,
+    ) -> Result<Proj, ProjCreateError> {
         let ctx = unsafe { std::mem::replace(&mut self.ctx, proj_context_create()) };
         transform_epsg(ctx, from, to, area)
     }
@@ -420,7 +462,29 @@ impl Default for ProjBuilder {
     }
 }
 
-/// A coordinate transformation object
+/// A coordinate transformation object.
+///
+/// A `Proj` can be constructed a few different ways:
+///
+/// * [`ProjBuilder`]
+/// * [`Proj::new`]
+/// * [`Proj::new_known_crs`]
+///
+/// # Examples
+///
+/// ```rust
+/// # use approx::assert_relative_eq;
+/// use proj::{Proj, Coord};
+///
+/// let from = "EPSG:2230";
+/// let to = "EPSG:26946";
+/// let nad_ft_to_m = Proj::new_known_crs(&from, &to, None).unwrap();
+/// let result = nad_ft_to_m
+///     .convert((4760096.421921f64, 3744293.729449f64))
+///     .unwrap();
+/// assert_relative_eq!(result.x(), 1450880.29, epsilon=1.0e-2);
+/// assert_relative_eq!(result.y(), 1141263.01, epsilon=1.0e-2);
+/// ```
 pub struct Proj {
     c_proj: *mut PJconsts,
     ctx: *mut PJ_CONTEXT,
@@ -437,13 +501,34 @@ impl Proj {
     /// For conversion operations, `definition` defines input, output, and
     /// any intermediate steps that are required. See the `convert` example for more details.
     ///
+    /// # Examples
+    ///
+    /// Construcing a `Proj` from a PROJ string definition:
+    ///
+    /// ```
+    /// let transformer = proj::Proj::new(
+    ///     "+proj=merc +lat_ts=56.5 +ellps=GRS80"
+    /// ).unwrap();
+    /// ```
+    ///
+    /// A `TryFrom` implementation is available which wraps `new`:
+    ///
+    /// ```
+    /// use std::convert::TryFrom;
+    ///
+    /// let transformer = proj::Proj::try_from(
+    ///     "+proj=merc +lat_ts=56.5 +ellps=GRS80"
+    /// ).unwrap();
+    /// ```
+    ///
     /// # Safety
+    ///
     /// This method contains unsafe code.
     // In contrast to proj v4.x, the type of transformation
     // is signalled by the choice of enum used as input to the PJ_COORD union
     // PJ_LP signals projection of geodetic coordinates, with output being PJ_XY
     // and vice versa, or using PJ_XY for conversion operations
-    pub fn new(definition: &str) -> Option<Proj> {
+    pub fn new(definition: &str) -> Result<Proj, ProjCreateError> {
         let ctx = unsafe { proj_context_create() };
         transform_string(ctx, definition)
     }
@@ -468,25 +553,38 @@ impl Proj {
     /// [remember](https://proj.org/development/reference/functions.html#c.proj_create_crs_to_crs)
     /// to reverse the coordinates of `Point` or `Coordinate` structs in order for a conversion operation to
     /// return correct results.
+    //
+    /// # Examples
     ///
-    ///```rust
-    /// # use approx::assert_relative_eq;
-    /// extern crate proj;
-    /// use proj::{Proj, Coord};
+    /// Constructing a `Proj` from a source CRS and target CRS:
     ///
-    /// let from = "EPSG:2230";
-    /// let to = "EPSG:26946";
-    /// let nad_ft_to_m = Proj::new_known_crs(&from, &to, None).unwrap();
-    /// let result = nad_ft_to_m
-    ///     .convert((4760096.421921f64, 3744293.729449f64))
-    ///     .unwrap();
-    /// assert_relative_eq!(result.x(), 1450880.29, epsilon=1.0e-2);
-    /// assert_relative_eq!(result.y(), 1141263.01, epsilon=1.0e-2);
+    /// ```rust
+    /// let transformer = proj::Proj::new_known_crs(
+    ///     "EPSG:2230",
+    ///     "EPSG:26946",
+    ///     None
+    /// ).unwrap();
+    /// ```
+    ///
+    /// A `TryFrom` implementation is available which wraps `new_known_crs`:
+    ///
+    /// ```rust
+    /// use std::convert::TryFrom;
+    ///
+    /// let transformer = proj::Proj::try_from((
+    ///     "EPSG:2230",
+    ///     "EPSG:26946"
+    /// )).unwrap();
     /// ```
     ///
     /// # Safety
+    ///
     /// This method contains unsafe code.
-    pub fn new_known_crs(from: &str, to: &str, area: Option<Area>) -> Option<Proj> {
+    pub fn new_known_crs(
+        from: &str,
+        to: &str,
+        area: Option<Area>,
+    ) -> Result<Proj, ProjCreateError> {
         let ctx = unsafe { proj_context_create() };
         transform_epsg(ctx, from, to, area)
     }
@@ -555,6 +653,9 @@ impl Proj {
                 }
             };
 
+            // comparing against float point sentinel values is a reasonable usage of exact
+            // floating point comparison
+            #[allow(clippy::float_cmp)]
             let area = if west != -1000.0 && south != -1000.0 && east != -1000.0 && north != -1000.0
             {
                 Some(Area {
@@ -861,6 +962,45 @@ impl Proj {
     }
 }
 
+impl convert::TryFrom<&str> for Proj {
+    type Error = ProjCreateError;
+
+    /// Create a `Proj` from a PROJ string definition.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use std::convert::TryFrom;
+    ///
+    /// let transformer = proj::Proj::try_from(
+    ///     "+proj=merc +lat_ts=56.5 +ellps=GRS80"
+    /// ).unwrap();
+    /// ```
+    fn try_from(definition: &str) -> Result<Proj, Self::Error> {
+        Proj::new(definition)
+    }
+}
+
+impl convert::TryFrom<(&str, &str)> for Proj {
+    type Error = ProjCreateError;
+
+    /// Create a `Proj` from a source CRS and target CRS.
+    ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// use std::convert::TryFrom;
+    ///
+    /// let transformer = proj::Proj::try_from((
+    ///     "EPSG:2230",
+    ///     "EPSG:26946"
+    /// )).unwrap();
+    /// ```
+    fn try_from((source_crs, target_crs): (&str, &str)) -> Result<Proj, Self::Error> {
+        Proj::new_known_crs(source_crs, target_crs, None)
+    }
+}
+
 struct PjInfo {
     id: Option<String>,
     description: Option<String>,
@@ -908,8 +1048,6 @@ impl Drop for ProjBuilder {
 
 #[cfg(test)]
 mod test {
-    use crate::proj;
-
     use super::*;
 
     #[derive(Debug)]
@@ -946,8 +1084,8 @@ mod test {
         // ETRS89
         let to = "EPSG:4258";
 
-        let online_builder = ProjBuilder::new();
-        let offline_builder = ProjBuilder::new();
+        let mut online_builder = ProjBuilder::new();
+        let mut offline_builder = ProjBuilder::new();
 
         assert_eq!(online_builder.network_enabled(), false);
         assert_eq!(offline_builder.network_enabled(), false);
@@ -1007,7 +1145,7 @@ mod test {
     #[should_panic]
     // This failure is a bug in libproj
     fn test_searchpath() {
-        let tf = ProjBuilder::new();
+        let mut tf = ProjBuilder::new();
         tf.set_search_paths(&"/foo").unwrap();
         let ipath = tf.info().unwrap().searchpath;
         let pathsep = if cfg!(windows) { ";" } else { ":" };
@@ -1018,7 +1156,7 @@ mod test {
     fn test_set_endpoint() {
         let from = "EPSG:4326";
         let to = "EPSG:4326+3855";
-        let tf = ProjBuilder::new();
+        let mut tf = ProjBuilder::new();
         let ep = tf.get_url_endpoint().unwrap();
         assert_eq!(&ep, "https://cdn.proj.org");
         tf.set_url_endpoint("https://github.com/georust").unwrap();
@@ -1038,6 +1176,33 @@ mod test {
         assert_relative_eq!(t.x(), 1450880.2910605022);
         assert_relative_eq!(t.y(), 1141263.0111604782);
     }
+
+    #[test]
+    fn test_from_crs_nul_error() {
+        match Proj::new_known_crs("\0", "EPSG:4326", None) {
+            Err(ProjCreateError::ArgumentNulError(_)) => (),
+            _ => unreachable!(),
+        }
+
+        match Proj::new_known_crs("EPSG:4326", "\0", None) {
+            Err(ProjCreateError::ArgumentNulError(_)) => (),
+            _ => unreachable!(),
+        }
+    }
+
+    #[test]
+    fn test_from_crs_error() {
+        match Proj::new_known_crs("EPSG:4326", "🦀", None) {
+            Err(ProjCreateError::ProjError(..)) => (),
+            _ => unreachable!(),
+        }
+
+        match Proj::new_known_crs("🦀", "EPSG:4326", None) {
+            Err(ProjCreateError::ProjError(..)) => (),
+            _ => unreachable!(),
+        }
+    }
+
     #[test]
     // Carry out a projection from geodetic coordinates
     fn test_projection() {
@@ -1106,11 +1271,25 @@ mod test {
         assert_relative_eq!(t.x(), 1450880.2910605022);
         assert_relative_eq!(t.y(), 1141263.0111604782);
     }
+
     #[test]
     // Test that instantiation fails wth bad proj string input
     fn test_init_error() {
-        assert!(Proj::new("🦀").is_none());
+        match Proj::new("🦀") {
+            Err(ProjCreateError::ProjError(_)) => (),
+            _ => unreachable!(),
+        }
     }
+
+    #[test]
+    // Test that instantiation fails wth bad proj string input
+    fn test_init_error_nul() {
+        match Proj::new("\0") {
+            Err(ProjCreateError::ArgumentNulError(_)) => (),
+            _ => unreachable!(),
+        }
+    }
+
     #[test]
     fn test_conversion_error() {
         // because step 1 isn't an inverse conversion, it's expecting lon lat input
